const { args } = require('./helpers')
const Lines = require('../lines')

class Generator {
  constructor ({ match, apply, uuid, index, km }) {
    this.match = match
    this._apply = apply
    this.uuid = uuid
    this.index = index
    this.km = km
  }

  getAPI (config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  toString () {
    return `Generator(${this.match}, ${this._apply})`
  }

  matches (objects, context, hierarchy, config) {
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    return this.match({ args: args(context, hierarchy), objects, global: objects, hierarchy, context, api: this.getAPI(config) })
  }

  apply (objects, g, gs, context, hierarchy, config, response, log) {
    if (!log) {
      throw "generators.apply argument log is required"
    }

    this.getAPI(config)
    let n = (id) => id
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id)
    }
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    const km = (name) => config.getConfig(name)
    return this._apply({ km, args: args(context, hierarchy), objects, log, global: objects, g, n, hierarchy, context, uuid: this.uuid, gs, config, response, api: this.getAPI(config) })
  }
}

class Generators {
  constructor (generators) {
    this.generators = generators
    this.logs = []
  };

  // assumed - properties added to contexts before the generators are called. For setting paraphrase property
  apply (args, contexts, assumed = {}) {
    let config = args.config
    let objects = args.objects
    let hierarchy = args.hierarchy
    let response = args.response
    if (Array.isArray(contexts)) {
    } else if (typeof contexts === 'object') {
      contexts = [contexts]
    } else {
      return String(contexts)
    }

    const contexts_prime = []
    for (const icontext in contexts) {
      let context = contexts[icontext]
      context = Object.assign({}, context, assumed)
      // let context_prime = JSON.stringify(context);
      let context_prime = context
      let applied = false;
      for (let igenerator = 0; igenerator < this.generators.length; ++igenerator) {
        const generator = this.generators[igenerator]
        if (generator.matches(objects, context, hierarchy, config)) {
          const g = (context) => this.apply(args, context, assumed)
          const log = (message) => { this.logs.push(message) }
          // this.logs.push(`Generators: applied ${generator.toString()}\n to\n ${JSON.stringify(context)}`)
          context_prime = generator.apply(objects, g, args.gs, context, hierarchy, config, response, log)
          if (!context_prime) {
            const message = `ERROR while applying ${generator.toString()}\n to\n ${JSON.stringify(context)}. The apply function did not return a value'`
            this.logs.push(message)
            throw { error: [message], logs: this.logs }
          }
          {
            const widths = [10, 10, 90]
            const lines = new Lines(widths)
            lines.setElement(0, 0, 'Generator')
            if (generator.index > -1 && generator.km) {
              lines.setElement(0, 2, `KM '${generator.km}' ordinal: ${generator.index}`)
            }
            lines.newRow()
            lines.setElement(0, 1, 'APPLIED')
            lines.setElement(0, 2, generator.toString())
            lines.newRow()
            lines.setElement(0, 1, 'TO')
            lines.setElement(0, 2, JSON.stringify(context, null, 2))
            lines.newRow()
            lines.setElement(0, 1, 'RESULT')
            lines.setElement(0, 2, context_prime)
            this.logs.push(lines.toString())
          }
          applied = true;
          break
        }
      }
      if (!applied) {
        const widths = [10, 10, 90]
        const lines = new Lines(widths)
        lines.setElement(0, 0, 'Generator')
        lines.setElement(0, 2, 'No generator applied')
        lines.newRow()
        lines.setElement(0, 1, 'TO')
        lines.setElement(0, 2, JSON.stringify(context, null, 2))
        this.logs.push(lines.toString())
      }
      contexts_prime.push(context_prime)
    }
    return contexts_prime
  }
}

module.exports = { Generator, Generators }
