const { args : contextArgs } = require('./helpers')
const Lines = require('../lines')

class Semantic {
  constructor ({match, apply, uuid, index, km}) {
    this.matcher = match
    this._apply = apply
    this.uuid = uuid
    this.index = index
    this.km = km
  }

  toString () {
    return `Semantic(${this.matcher}, ${this._apply})`
  }

  getAPI (config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  matches (args, context) {
    const hierarchy = args.hierarchy
    const config = args.config
    const objects = args.getObjects(this.uuid)
    return this.matcher({ args: contextArgs(context, hierarchy), objects, global: objects, context, hierarchy, api: this.getAPI(config) })
  }

  apply (args, context, s, log) {
    const { hierarchy, config, response } = args;
    const objects = args.getObjects(this.uuid)
    if (!log) {
      console.trace()
      throw 'log is a required argument'
    }
    const context_prime = Object.assign({}, context)
    let n = (id) => id
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id)
    }
    const km = (name) => config.getConfig(name)
    this._apply({ km, args: contextArgs(context, hierarchy), objects, log, global: objects, n, hierarchy, context: context_prime, uuid: this.uuid, config, response, s, api: this.getAPI(config) })
    return context_prime
  }
}

class Semantics {
  constructor (semantics) {
    this.semantics = semantics
    this.logs = []
  };

  applyToContext (args, context) {
    const { objects, hierarchy, config, response } = args;
    //let context_prime = {}
    if (!(context instanceof Array || context instanceof Object)) {
      return context
    }

    let context_prime = Object.assign({}, context)
    const s = (context) => this.apply(args, context)
    let applied = false;
    for (const isemantic in this.semantics) {
      const semantic = this.semantics[isemantic]
      if (semantic.matches(args, context)) {
        const log = (message) => { this.logs.push(message) }
        context_prime = semantic.apply(args, context, s, log)
        //this.logs.push(`Semantics: applied ${semantic.toString()}\n to\n ${JSON.stringify(context)}\n the result was ${JSON.stringify(context_prime)}\n`)
        {
          const widths = [10, 10, 90]
          const lines = new Lines(widths)
          lines.setElement(0, 0, 'Semantic')
          if (semantic.index > -1 && semantic.km) {
            lines.setElement(0, 2, `KM '${semantic.km}' ordinal: ${semantic.index}`)
          }
          lines.newRow()
          lines.setElement(0, 1, 'APPLIED')
          lines.setElement(0, 2, semantic.toString())
          lines.newRow()
          lines.setElement(0, 1, 'TO')
          lines.setElement(0, 2, JSON.stringify(context, null, 2))
          lines.newRow()
          lines.setElement(0, 1, 'RESULT')
          lines.setElement(0, 2, JSON.stringify(context_prime, null, 2))
          this.logs.push(lines.toString())
        }
        applied = true;
        break
      }
    }
    if (!applied) {
      const widths = [10, 10, 90]
      const lines = new Lines(widths)
      lines.setElement(0, 0, 'Semantic')
      lines.setElement(0, 2, 'No semantic applied')
      lines.newRow()
      lines.setElement(0, 1, 'TO')
      lines.setElement(0, 2, JSON.stringify(context, null, 2))
      this.logs.push(lines.toString())
    }
    return context_prime
  }

  applyToContexts (args, contexts) {
    const contexts_prime = []
    contexts.forEach((context) => {
      contexts_prime.push(this.applyToContext(args, context))
    })
    return contexts_prime
  }

  apply (args, context) {
    if (Array.isArray(context)) {
      return this.applyToContexts(args, context)
    } else if (context instanceof Object) {
      return this.applyToContext(args, context)
    } else {
      return context
    }
  }
}

module.exports = { Semantic, Semantics }
